#include <stdlib.h>

#include "list.h"

struct list_node
   {
   struct list_node * next;
   struct list_node * prev;
   void * data;
   };

struct list
   {
   struct list_node * head;
   struct list_node * tail;
   size_t size;
   };

struct list_iterator
   {
   struct list * l;
   struct list_node * curr;
   };

////////////////////////////////////////////////////////////////////////////////
// Allocates new list and returns poitner to it.
struct list *
list_new(void)
   {
   struct list * l = malloc(sizeof(struct list));
   if (NULL == l)
      return NULL;

   l->size = 0;
   l->head = l->tail = NULL;

   return l;
   }

////////////////////////////////////////////////////////////////////////////////
// Deletes list.  Does not delete data elements.  To avoid memory leaks,
// Only call this on an empty list.
void
list_delete(struct list * l)
   {
   if (NULL != l)
      {
      // This will cause memory leaks, but we don't have destructors.
      // Hopefuly the caller has already emptied the list.
      struct list_node * prev;
      struct list_node * curr = l->tail;
      while (curr)
	 {
	 prev = curr->prev;
	 free(curr);
	 curr = prev;
	 }
      free(l);
      }
   }

////////////////////////////////////////////////////////////////////////////////
// Returns the numb er of elements in the list.
size_t
list_size(struct list * l)
   {
   if (NULL == l)
      return 0;
   return l->size;
   }

////////////////////////////////////////////////////////////////////////////////
// Add an element to the end of the list.
void *
list_push_back(struct list * l, void * elem)
   {
   if ( (NULL == l) || (NULL == elem) )
      return NULL;

   struct list_node * new = malloc(sizeof(struct list_node));
   if (NULL == new)
      return NULL;

   new->data = elem;
   new->prev = new->next = NULL;

   if (NULL == l->tail)
      {
      l->head = l->tail = new;
      }
   else
      {
      new->prev = l->tail;
      l->tail->next = new;
      l->tail = new;
      }

   l->size++;

   return elem;
   }

////////////////////////////////////////////////////////////////////////////////
// Remove an element from the back of the list.
void *
list_pop_back(struct list * l)
   {
   if ( (NULL == l) || (NULL == l->tail) )
      return NULL;
   
   void * t = l->tail->data;

   struct list_node * prev = l->tail->prev;
   if (prev)
      prev->next = NULL;
   free(l->tail);
   l->tail = prev;
   l->size--;
   if (0 == l->size)
      l->head = NULL;

   return t;
   }

////////////////////////////////////////////////////////////////////////////////
// Add an element to the front of the list.
void *
list_push_front(struct list * l, void * elem)
   {
   if ( (NULL == l) || (NULL == elem) )
      return NULL;

   struct list_node * new = malloc(sizeof(struct list_node));
   if (NULL == new)
      return NULL;

   new->data = elem;
   new->prev = new->next = NULL;

   if (NULL == l->head)
      {
      l->head = l->tail = new;
      }
   else
      {
      new->next = l->head;
      l->head->prev = new;
      l->head = new;
      }

   l->size++;

   return elem;
   }

////////////////////////////////////////////////////////////////////////////////
// Remove and element from the front of the list.
void *
list_pop_front(struct list * l)
   {
   if ( (NULL == l) || (NULL == l->head) )
      return NULL;
   
   void * t = l->head->data;

   struct list_node * next = l->head->next;
   if (next)
      next->prev = NULL;
   free(l->head);
   l->head = next;
   l->size--;

   if (0 == l->size)
      l->tail = NULL;

   return t;
   }





////////////////////////////////////////////////////////////////////////////////
// Intialize an iterator to point to the first element of the list
struct list_iterator *
list_begin(struct list * l)
   {
   if ( (NULL == l) || (NULL == l->head) )
      return NULL;

   struct list_iterator * it = malloc(sizeof(struct list_iterator));
   if (NULL == it)
      return NULL;

   it->l = l;
   it->curr = l->head;

   return it;
   }

////////////////////////////////////////////////////////////////////////////////
// Intialize an iterator to point to the last element of the list
struct list_iterator *
list_end(struct list * l)
   {
   if ( (NULL == l) || (NULL == l->head) )
      return NULL;

   struct list_iterator * it = malloc(sizeof(struct list_iterator));
   if (NULL == it)
      return NULL;

   it->l = l;
   it->curr = l->tail;

   return it;
   }

////////////////////////////////////////////////////////////////////////////////
// Move the iterator to the next element in the list.
struct list_iterator *
list_next(struct list_iterator * it)
   {
   if (NULL == it)
      return NULL;

   it->curr = it->curr->next;
   return it->curr;
   }

////////////////////////////////////////////////////////////////////////////////
// Move the iterator to the previous element in the list.
struct list_iterator *
list_prev(struct list_iterator * it)
   {
   if (NULL == it)
      return NULL;

   it->curr = it->curr->prev;
   return it->curr;
   }

////////////////////////////////////////////////////////////////////////////////
// Remove the current element pointed to by the iterator.
// Returns the data element that was removed so the caller can free it.
void *
list_remove(struct list_iterator * it)
   {
   if ( (NULL == it) || (NULL == it->curr) )
      return NULL;

   if (it->curr->prev)
      it->curr->prev->next = it->curr->next;

   if (it->curr->next)
      it->curr->next->prev = it->curr->prev;

   struct list_node * temp = it->curr;
   void * elem = it->curr->data;

   it->curr = it->curr->next;
   free(temp);

   it->l->size--;
   if (0 == it->l->size)
      it->l->tail = it->l->head = NULL;

   return elem;
   }

////////////////////////////////////////////////////////////////////////////////
// Return the value of the element pointed to by the iterator.
void *
list_value(struct list_iterator * it)
   {
   if ( (NULL == it) || (NULL == it->curr) )
      return NULL;
   return it->curr->data;
   }

////////////////////////////////////////////////////////////////////////////////
// Insert a new element after the one pointed to by the iterator.  
// The iterator is then set to point to this new element.
struct list_iterator *
list_insert_after(struct list_iterator * it, void * elem)
   {
   if ( (NULL == it) || (NULL == elem) || (NULL == it->curr) )
      return NULL;

   struct list_node * new = malloc(sizeof(struct list_node));
   if (NULL == new)
      return NULL;

   new->data = elem;
   new->prev = it->curr;
   if (it->curr->next)
      it->curr->next->prev = new;
   new->next = it->curr->next;
   it->curr->next = new;

   it->curr = new;

   it->l->size++;
   if (NULL == it->curr->next)
      it->l->tail = it->curr;

   return it;
   }

////////////////////////////////////////////////////////////////////////////////
// Insert a new element before the one pointed to by the iterator.  
// The iterator is then set to point to this new element.
struct list_iterator *
list_insert_before(struct list_iterator * it, void * elem)
   {
   if ( (NULL == it) || (NULL == elem) || (NULL == it->curr) )
      return NULL;

   struct list_node * new = malloc(sizeof(struct list_node));
   if (NULL == new)
      return NULL;

   new->data = elem;
   new->next = it->curr;
   if (it->curr->prev)
      it->curr->prev->next = new;
   new->prev = it->curr->prev;
   it->curr->prev = new;

   it->curr = new;

   it->l->size++;
   if (NULL == it->curr->prev)
      it->l->head = it->curr;

   return it;
   }

